#include <stdio.h>
#include <string.h>
#include <time.h>

#include "q3la.h"

q3la_main::q3la_main()
{
	gamecount = 0;
	logfilecount = 0;
//	maps = new map_t(&config);
}

q3la_main::~q3la_main()
{
//	delete maps;
}

void q3la_main::run (int argc, char** argv) 
{
#ifdef TEST
	_chdir (".");
	argc = 2;
	argv[1] = "server.log";
#endif
	
	if (argc < 2) {
		help();
	}
	
	version_info();

	// get start time
	
	clock_t start, finish;   
	start = clock();
	
	config.read_options (argc, argv); // read command line options - needed for read_cfg
	printf ("\nReading CFG: %s", config.configfile);
	config.read_cfg (config.configfile); // read CFG
	printf ("\n");
	config.read_options (argc, argv); // read command line options again to override CFG
	
#ifdef TEST
//	config.limits_used = true;
//	config.max_players = 20;
//	config.sortorder_players = so_playtime;
	config.top_players = 20;
	config.chatlog = true;
	config.no_bots = false;
	strcpy (config.output_directory, "c:\\results\\");
#endif
	
	if (config.q3ladata == true) { // delete q3ladata.txt
		outputfile deldata; 
		deldata.open("q3ladata.txt");
		deldata.close();
	}
	config.print_settings();
#ifndef __LINUX_PLATFORM__
	analyze (config.filespec);
#else
	analyze (argc,argv);
#endif
	
	// Get finish time, output
	
	finish = clock();
	printf( "\n%.2f seconds elapsed", ((double) (finish - start) / CLOCKS_PER_SEC));
	
	if (config.open_after == true) { // NT / 9x check!!!
		system ("start players.htm"); //9x
		system ("players.htm"); // NT
	}
	
	//	config.item_list(); // TEST
}

//
// ANALYZE
//

#ifndef __LINUX_PLATFORM__
void q3la_main::analyze (char* pathstr) 
#else
void q3la_main::analyze (int argc, char** argv)
#endif
{
	resultdata_total results_total(&config);
	
	// Backup start directory

#ifndef __LINUX_PLATFORM__
	_getcwd (start_directory, 256);
#else
	getcwd (start_directory, 256);
#endif	
	
	// Try to read CSS file if output directory not == current directory
	
	int css_in = 0;
	int css_size = 0;
	char* css_buf = 0;
	
	if ((config.output_directory[0] != 0) && (strcmp (config.output_directory, ".") != 0)) {
#ifndef __LINUX_PLATFORM__
		css_in = _open ("q3la.css", _O_BINARY);
#else
		css_in = open ("q3la.css", O_RDONLY);
#endif
		if (css_in != -1) {
			config.q3la_css = true;
		}
		css_size = _filelength (css_in);
		css_buf = new char[css_size+1];
		printf ("Reading q3la.css\n");
		_read (css_in, css_buf, css_size);
		_close (css_in);
	}
	
	// Create and change to output directory
	
	_mkdir (config.output_directory);
	_chdir (config.output_directory);
	
	// Delete directory "q3la_games"
	
	diverse::deltree ("q3la_games");
	diverse::deltree ("q3la_players");
	
	// Write CSS file (if it was read and not already exists in results folder)
	
	if (config.q3la_css == true) { // CSS file was read
		FILE* css_out = fopen ("q3la.css", "r");
		if (css_out == 0) {  // CSS doesn't yet exist
			FILE* css_out = fopen ("q3la.css", "wb");
			if (css_out == 0) { // failed to copy
				printf ("\nCouldn't write q3la.css, using default stylesheets");
				config.q3la_css = false;
			} else { // copy successful
				printf ("\nCopied stylesheet file q3la.css to results folder");
				fwrite (css_buf, sizeof(char), css_size, css_out);
				fclose (css_out);
			}
		} else { // CSS already exists
			printf ("\nUsing existing stylesheet file q3la.css");
			fclose (css_out);
		}
	} else { // no CSS file read
		FILE* css_out = fopen ("q3la.css", "r");
		if (css_out != 0) { // CSS exists
			printf ("\nUsing existing stylesheet file q3la.css");
			fclose (css_out);
			config.q3la_css = true;
		} else { // CSS doesn't exist
			printf ("\nNo q3la.css found, using default stylesheets");
		}
	}
	printf ("\n");
	
	delete[] css_buf;
	
	// Change back to start directory to find logs
	
	_chdir (start_directory);
	
	// Start parsing logs
	
#ifndef __LINUX_PLATFORM__
	_finddata_t finddata;
	int handle = _findfirst (pathstr, &finddata);
	int done = handle;
	while (done != -1) {
		if ((finddata.attrib != _A_SUBDIR) && (finddata.name[0] != '.')) {
			logfilecount++;			
			analyze_one_file(finddata.name, &results_total);
		};
		
		// Change back to start directory for reading next logfile (necesssary???)
		
		_chdir (start_directory);
		
		// Get next logfile 
		
		done = _findnext (handle, &finddata);
	};
#else
	for(int j=1;j<argc;j++){
		char *pathstr=argv[j];
		if (*pathstr!='-' && *pathstr!='/'){
			logfilecount++;			
			analyze_one_file(pathstr, &results_total);
		};
		_chdir (start_directory);
	};
#endif
	
	// If no logs were found, exit

	if (logfilecount == 0) {
		error::msg_and_exit ("No logfiles found");
	}
	
	printf ("\n");

	// If games<x>.htm still open, write footer and close
	
	if (out_games_total.is_open()) {
		games_footer();
	}
	
	// Change to output directory for writing total results
	
	_chdir (config.output_directory);
	
	// Write game index (or rename games1.htm to games.htm, if <= config.game_split games)
	
	game_index();
	
	// Change to output directory for writing player stats
	
	_chdir (start_directory);
	_chdir (config.output_directory);

	// Write player stats and top player stats

	output_player_stats(&results_total, 0);
	if (config.top_players > 0) {
		output_player_stats(&results_total, config.top_players);
	}
	
	// Output per player results
	
#ifndef __LINUX_PLATFORM__
	_mkdir ("q3la_players");
#else
	mkdir ("q3la_players", 0755);
#endif
	_chdir ("q3la_players");
	config.in_subdir = true;
	
	printf ("\nWriting per player stats        ");
	output_per_player_stats(&results_total);
	printf ("\n");

	// Map stats

/*
	map_t* temp_map = maps->next;
	while (temp_map != 0) {
		printf ("\nMap \"%s\"", temp_map->mapname);
		temp_map = temp_map->next;
	}

*/
	
	// Go back to start directory
	
	_chdir (start_directory);
	config.in_subdir = false;
}

void q3la_main::analyze_one_file(char *filename, resultdata_total *results_total){
	FILE* logfile = fopen (filename, "rb");
	if (logfile == 0) {
		error::msg_and_exit ("q3la_main::analyze - can't open %s", filename);
	}
	
	fseek (logfile, 0, SEEK_END);
	int logfilesize = ftell (logfile);
	fseek (logfile, 0, SEEK_SET);
	
	printf ("\nParsing %s [%i byte]", filename, logfilesize);
	
	// Change to output directory
	
	_chdir (config.output_directory);
	
	// Write stats
	
	printf ("\n- Writing game stats        ");
	
	resultdata result(&config);
	analyzelog analyze (logfile, &result, &config);
	for (;;) {
		result.init();
		if (analyze.parse() == 0) { // parse next game
			break;
		}
		result.sort_players(config.sortorder_players_2, false); // sort players for each game
		result.sort_players(config.sortorder_players_1, false); // sort players for each game
		
		int rel_pos = 100 * ftell(logfile) / logfilesize;
		printf ("\b\b\b\b\b\b\b(%3i %%)", rel_pos); // progress indication
#ifdef __LINUX_PLATFORM__
		fflush(stdout);
#endif
		
		if (output_game(&result, &analyze) == 0) {
			// save game data
			_chdir (start_directory);
			if (config.q3ladata) result.save("q3ladata.txt");
			_chdir (config.output_directory);
			//-- save game data
			results_total->add_results (&result); // add to total results 
			// maps
//			map_t* temp_map = maps->get_map (result.mapname);
//			temp_map->result->add_results (&result);
		}
	}
	printf ("\b\b\b\b\b\b\b(100 %%)"); // progress indication
#ifdef __LINUX_PLATFORM__
	fflush(stdout);
#endif
	
	if (config.q3ladata == true) { 
		printf ("\nWriting q3ladata.txt"); // output of this is delayed to actual writing
	}

	fclose (logfile);
}

//
// PER PLAYER STATS
//

void q3la_main::output_per_player_stats(resultdata_total* results_total)
{
	player_t* player = results_total->first_player.next;
	int count = 0;
	while (player != 0) {
		count++;
		if (player->listed == false) {
			player = player->next;
			continue;
		}
		int rel_pos = 100 * count / results_total->get_player_count();
		printf ("\b\b\b\b\b\b\b(%3i %%)", rel_pos); // progress indication
#ifdef __LINUX_PLATFORM__
		fflush(stdout);
#endif
		outputfile out_player;
		char filename[256];
		sprintf (filename, "player_%i.htm", player->iterator);
		out_player.open (filename);
		html::html_header (&out_player, &config);
		out_player.write ("<p class=\"title\"><a href=\"../players.htm\"><font color=red>All players</font></a></p>");
		if (config.top_players > 0) {
			out_player.write ("<p class=\"title\"><a href=\"../top_players.htm\"><font color=red>Top %i players</font></a></p>", config.top_players);
		}
		out_player.write ("<p class=\"title\"><span style=\"background-color: #808080\">%s</span></p>", 
			player->get_htmlname(config.no_funnames));
		output out (results_total, &out_player, &config, 0);
		out.player_info(player);
		out.player_weapon_stats(player);
		out.player_to_player_stats(player);
		html::html_footer (&out_player, false);
		out_player.close();
		player = player->next;
	}
	printf ("\b\b\b\b\b\b\b(100 %%)"); // progress indication
#ifdef __LINUX_PLATFORM__
	fflush(stdout);
#endif
}

//
// PER GAME STATS
//

int q3la_main::output_game(resultdata* result, analyzelog* analyze)
{
	// delete bots
	if (config.no_bots == true) {
		result->delete_bots();
	}
	
	// delete nullplayers
	result->delete_nullplayers();
	
	// only output games with more than one player
	if (result->get_player_count() < 2) {
		return 1;
	}
	
	gamecount++;

	if (config.chatlog) {
		analyze->chatlog.write ("<p><i><b>Game %i</b></i></p>\r\n", gamecount+1);	
	}

#ifndef __LINUX_PLATFORM__
	_mkdir ("q3la_games");
#else
	mkdir ("q3la_games", 0755);
#endif
	
	// Open games<x>.htm?
	if (gamecount % config.game_split == 1) {
		games_header();
	}
	
	// Output to "games<x>.htm"
	output out_total (result, &out_games_total, &config);
	html::line_header(&out_games_total);
	out_games_total.write ("<b><font size=\"3\"><a href=\"game_%i.htm\"><font color=\"#FF0000\">Game %i</font></a><br>(%s - %s)</font></b>", gamecount, gamecount, result->mapname, result->gametype);
	html::line_footer(&out_games_total);
	out_total.list_teams_html();
	out_total.list_players_html(false); // !!!
	
	// Open output file for current game
	
	_chdir ("q3la_games");
	config.in_subdir = true;
	
	char gamefile[16];
	sprintf (gamefile, "game_%i.htm", gamecount);
	outputfile out_game_current;
	out_game_current.open (gamefile);
	html::html_header(&out_game_current, &config);
	
	// Output to "game_<x>.htm"
	output out_current (result, &(out_game_current), &config);
	
#ifndef __LINUX_PLATFORM__
	// output creation time
	
    struct tm *newtime;
    time_t long_time;

    time( &long_time );                // get time as long integer
    newtime = localtime( &long_time ); // convert to local time

    out_game_current.write ("<p class=text>[%s]</p>\r\n", asctime(newtime)); 

	//-- output creation time
#endif

	if (config.homepage_address[0] != 0) {
		out_game_current.write ("<p class=\"title\"><a href=\"%s\"><font color=red>%s</font></a></p>", 
			config.homepage_address, config.homepage_string);
	}
	out_game_current.write ("<p class=\"title\"><a href=\"../players.htm\"><font color=red>All players</font></a></p>");
	if (config.top_players > 0) {
		out_game_current.write ("<p class=\"title\"><a href=\"../top_players.htm\"><font color=red>Top %i players</font></a></p>", config.top_players);
	}
	out_game_current.write ("<p class=\"title\"><a href=\"../games.htm\"><font color=red>Game index</font></a></p>");
	out_game_current.write ("<p class=\"title\"><a href=\"games%i.htm\"><font color=red>Game list</font></a></p>", (gamecount-1) / config.game_split + 1);
	out_game_current.write ("<p class=\"title\">Game %i</p>", gamecount);
	out_current.game_info();
	out_current.list_teams_html();
	out_current.list_players_html(true);
	out_current.list_health_stats_html();
	out_current.list_powerup_stats_html();
	out_current.list_weapon_stats_html();
	out_current.list_ammo_stats_html();
	html::line (&out_game_current, "<b>Killcharts</b>");
	for (int i = 0; i < MAX_WEAPONS; i++) {
		out_current.table_kills_html(i);
	}
	if (config.q3ladata == true) {
		out_current.game_data(gamecount);
	}
	
	// Close current game file
	html::html_footer(&out_game_current, true);
	out_game_current.close();
	_chdir ("..");
	config.in_subdir = false;
	
	// Close games<x>.htm?
	if (gamecount % config.game_split == 0) {
		games_footer();
	}
	
	return 0;
}

void q3la_main::help () 
{
	version_info();
	config.print_help();
	printf ("\n\nOptions can be specified in upper or lower case with the prefix '/' or '-'");
	printf ("\n");
	exit (1);
}

void q3la_main::version_info()
{
    printf ("\n%s %s, %s\n%s\n", helpstr1, version, date, helpstr2);
}

void q3la_main::games_header()
{
	_chdir ("q3la_games");
	config.in_subdir = true;
	char filename[256];
	sprintf (filename, "games%i.htm", gamecount / config.game_split + 1);
	out_games_total.open (filename);
	html::html_header (&out_games_total, &config);

#ifndef __LINUX_PLATFORM__
	// output creation time
	
    struct tm *newtime;
    time_t long_time;

    time( &long_time );                // get time as long integer
    newtime = localtime( &long_time ); // convert to local time

    out_games_total.write ("<p class=text>[%s]</p>\r\n", asctime(newtime)); 

	//-- output creation time
#endif

	if (config.homepage_address[0] != 0) {
		out_games_total.write ("<p class=\"title\"><a href=\"%s\"><font color=red>%s</font></a></p>", 
			config.homepage_address, config.homepage_string);
	}
	out_games_total.write ("<p class=\"title\"><a href=\"../players.htm\"><font color=red>All players</font></a></p>");
	if (config.top_players > 0) {
		out_games_total.write ("<p class=\"title\"><a href=\"../top_players.htm\"><font color=red>Top %i players</font></a></p>", config.top_players);
	}
	out_games_total.write ("<p class=\"title\"><a href=\"../games.htm\"><font color=red>Game Index</font></a></p>");
	html::line (&out_games_total, "<b>Games with less than two players are not listed\r\n</b>");
	html::line (&out_games_total, "<b>Players without kills or deaths are ignored</b>");
	html::line (&out_games_total, "<b>Players are sorted by %s and %s</b><br>(in team games by team and score)", sortorder_string(config.sortorder_players_1), sortorder_string(config.sortorder_players_2));
	_chdir ("..");
	config.in_subdir = false;
}

void q3la_main::games_footer()
{
	html::html_footer(&out_games_total, true);
	out_games_total.close();
}

void q3la_main::game_index()
{
	outputfile outfile;
	outfile.open ("games.htm");
	
#ifndef __LINUX_PLATFORM__
	// output creation time
	
    struct tm *newtime;
    time_t long_time;

    time( &long_time );                // get time as long integer
    newtime = localtime( &long_time ); // convert to local time

    outfile.write ("<p class=text>[%s]</p>\r\n", asctime(newtime)); 

	//-- output creation time
#endif

	if (config.homepage_address[0] != 0) {
		outfile.write ("<p class=\"title\"><a href=\"%s\"><font color=red>%s</font></a></p>", 
			config.homepage_address, config.homepage_string);
	}
	outfile.write ("<p class=\"title\"><a href=\"players.htm\"><font color=red>All players</font></a></p>");
	if (config.top_players > 0) {
		outfile.write ("<p class=\"title\"><a href=\"top_players.htm\"><font color=red>Top %i players</font></a></p>", config.top_players);
	}
	
	html::html_header (&outfile, &config);
	for (int i = 0; i <= (gamecount - 1) / config.game_split; i++) {
		int num2;
		if (i != gamecount / config.game_split) {
			num2 = i*config.game_split+config.game_split;
		} else {
			num2 = gamecount;
		}
		outfile.write (
			"<p  class=\"title\"><a href=\"q3la_games/games%i.htm\"><font color=\"#FFFFFF\"><b>Games %i - %i</b></font></a></p>",
			i+1, i*config.game_split+1, num2);
	}
	html::html_footer (&outfile, false);
	outfile.close();
}

void q3la_main::output_player_stats(resultdata_total* results_total, int top)
{
	outputfile outfile;

	char outfilename[256];
	char top_str[16];
	if (top == 0) {
		strcpy (top_str, "");
	} else {
		strcpy (top_str, "top_");
	}

	// Output total player results, default
	
	sprintf (outfilename, "%s%s", top_str, "players.htm");
	outfile.open (outfilename);
	printf ("\nWriting player stats to \"%s\"", outfilename);
	
	results_total->sort_players (config.sortorder_players_2); // sort all players
	results_total->sort_players (config.sortorder_players_1); // sort all players
	output out_total (results_total, &outfile, &config, gamecount);
	out_total.list_players_total_html(top);
	
	outfile.close ();

	// Output total player results, by ratio
	
	sprintf (outfilename, "%s%s", top_str, "players_ratio.htm");
	outfile.open (outfilename);
	printf ("\nWriting player stats to \"%s\"", outfilename);
	
	results_total->sort_players (so_ratio); // sort all players
	output out_total1 (results_total, &outfile, &config, gamecount);
	out_total1.list_players_total_html(top);
	
	outfile.close ();
	
	// Output total player results, by name
	
	sprintf (outfilename, "%s%s", top_str, "players_name.htm");
	outfile.open (outfilename);
	printf ("\nWriting player stats to \"%s\"", outfilename);
	
	results_total->sort_players (so_name); // sort all players
	output out_total2 (results_total, &outfile, &config, gamecount);
	out_total2.list_players_total_html(top);
	
	outfile.close ();
	
	// Output total player results, by eff
	
	sprintf (outfilename, "%s%s", top_str, "players_eff.htm");
	outfile.open (outfilename);
	printf ("\nWriting player stats to \"%s\"", outfilename);
	
	results_total->sort_players (so_eff); // sort all players
	output out_total3 (results_total, &outfile, &config, gamecount);
	out_total3.list_players_total_html(top);
	
	outfile.close ();
	
	// Output total player results, by skill
	
	sprintf (outfilename, "%s%s", top_str, "players_skill.htm");
	outfile.open (outfilename);
	printf ("\nWriting player stats to \"%s\"", outfilename);
	
	results_total->sort_players (so_skill); // sort all players
	output out_total4 (results_total, &outfile, &config, gamecount);
	out_total4.list_players_total_html(top);
	
	outfile.close ();
	
	// Output total player results, by frags
	
	sprintf (outfilename, "%s%s", top_str, "players_frags.htm");
	outfile.open (outfilename);
	printf ("\nWriting player stats to \"%s\"", outfilename);
	
	results_total->sort_players (so_frags); // sort all players
	output out_total5 (results_total, &outfile, &config, gamecount);
	out_total5.list_players_total_html(top);
	
	outfile.close ();
	
	// Output total player results, by kills
	
	sprintf (outfilename, "%s%s", top_str, "players_kills.htm");
	outfile.open (outfilename);
	printf ("\nWriting player stats to \"%s\"", outfilename);
	
	results_total->sort_players (so_kills); // sort all players
	output out_total6 (results_total, &outfile, &config, gamecount);
	out_total6.list_players_total_html(top);
	
	outfile.close ();
	
	// Output total player results, by playtime
	
	sprintf (outfilename, "%s%s", top_str, "players_playtime.htm");
	outfile.open (outfilename);
	printf ("\nWriting player stats to \"%s\"", outfilename);
	
	results_total->sort_players (so_playtime); // sort all players
	output out_total7 (results_total, &outfile, &config, gamecount);
	out_total7.list_players_total_html(top);

	outfile.close ();
}
